# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines common private constants."""


class ManagerNames(object):
    """Provide the manager names."""

    CAUSAL = 'causal'
    COUNTERFACTUAL = 'counterfactual'
    DATA_BALANCE = 'data_balance'
    ERROR_ANALYSIS = 'error_analysis'
    EXPLAINER = 'explainer'


class Metadata(object):
    """Provide constants for metadata and saved files."""

    META_JSON = 'meta.json'
    MODEL = 'model'
    TRAIN = 'train'
    TEST = 'test'
    TASK_TYPE = 'task_type'
    TARGET_COLUMN = 'target_column'
    CLASSES = 'classes'
    FEATURE_COLUMNS = 'feature_columns'
    FEATURE_METADATA = 'feature_metadata'
    FEATURE_RANGES = 'feature_ranges'
    CATEGORICAL_FEATURES = 'categorical_features'
    CATEGORIES = 'categories'
    CATEGORY_DICTIONARY = 'category_dictionary'
    CATEGORICAL_INDEXES = 'categorical_indexes'
    STRING_IND_DATA = 'string_ind_data'
    NUMBER_LARGE_TEST_SAMPLES = 'number_large_test_samples'


class ListProperties(object):
    """Provide constants for listing manager properties."""

    MANAGER_TYPE = 'manager_type'


class ExplainerManagerKeys(object):
    """Provide constants for ExplainerManager key properties."""
    ID = 'id'
    IS_COMPUTED = 'is_computed'
    IS_ENGINEERED = 'is_engineered'
    IS_RAW = 'is_raw'
    METHOD = 'method'
    MODEL_TASK = 'model_task'
    MODEL_TYPE = 'model_type'


class ErrorAnalysisManagerKeys(object):
    """Provide constants for ErrorAnalysisManager key properties."""
    FILTER_FEATURES = 'filter_features'
    IS_COMPUTED = 'is_computed'
    MAX_DEPTH = 'max_depth'
    MIN_CHILD_SAMPLES = 'min_child_samples'
    NUM_LEAVES = 'num_leaves'
    REPORTS = 'reports'


class CounterfactualManagerKeys(object):
    """Provide constants for CounterfactualManager key properties."""
    COUNTERFACTUALS = 'counterfactuals'


class CausalManagerKeys(object):
    """Provide constants for CausalManager key properties."""
    CAUSAL_EFFECTS = 'causal_effects'
    GLOBAL_EFFECTS_COMPUTED = 'global_effects_computed'
    LOCAL_EFFECTS_COMPUTED = 'local_effects_computed'
    POLICIES_COMPUTED = 'policies_computed'


class DataBalanceManagerKeys(object):
    """Provide constants for DataBalanceManager key properties."""
    IS_ADDED = 'is_added'
    TASK_TYPE = 'task_type'
    COLS_OF_INTEREST = 'cols_of_interest'
    TARGET_COLUMN = 'target_column'
    CLASSES = 'classes'


class ExplanationKeys(object):

    EBM_GLOBAL_EXPLANATION_KEY = 'ebm_global'
    EXPLANATION_TYPE_KEY = 'explanation_type'
    GLOBAL_EXPLANATION_KEY = 'global_feature_importance'
    LOCAL_EXPLANATION_KEY = 'local_feature_importance'


class SerializationAttributes:
    """Constants for a serialized result."""

    # File structure
    RESULTS_DIRECTORY = 'results'
    PREDICTIONS_DIRECTORY = 'predictions'
    DATA_DIRECTORY = 'data'
    DASHBOARD_SCHEMAS = 'dashboard_schemas'

    # Metadata keys
    ID_KEY = 'id'
    VERSION_KEY = 'version'

    # Metadata filenames
    ID_FILENAME = 'id.json'
    VERSION_FILENAME = 'version.json'
    META_JSON = 'meta.json'
    RAI_VERSION_JSON = 'rai_version.json'

    # Dashboard filenames
    DASHBOARD_FILENAME = 'dashboard.json'

    # Model filenames
    MODEL_PKL = 'model.pkl'

    # Prediction filenames
    PREDICT_JSON = "predict.json"
    FORECAST_JSON = "forecast.json"
    PREDICT_PROBA_JSON = "predict_proba.json"
    FORECAST_QUANTILES_JSON = "forecast_quantiles.json"
    LARGE_PREDICT_JSON = "large_predict.json"
    LARGE_FORECAST_JSON = "large_forecast.json"
    LARGE_PREDICT_PROBA_JSON = "large_predict_proba.json"
    LARGE_FORECAST_QUANTILES_JSON = "large_forecast_quantiles.json"

    # Data filenames
    LARGE_TEST_JSON = "large_test.json"


class FileFormats:
    """Constants relating to file formats."""
    JSON = '.json'
    PKL = '.pkl'
    TXT = '.txt'



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the RAIBaseInsights class."""

import json
import pickle
import warnings
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Optional

import pandas as pd

import responsibleai
from raiutils.exceptions import UserConfigValidationException
from responsibleai._internal.constants import (FileFormats, Metadata,
                                               SerializationAttributes)

_DTYPES = 'dtypes'
_MODEL_PKL = Metadata.MODEL + FileFormats.PKL
_SERIALIZER = 'serializer'
_MANAGERS = 'managers'


class RAIBaseInsights(ABC):
    """Defines the base class RAIBaseInsights for the top-level API.

    This class is abstract and should not be instantiated.
    """

    def __init__(self, model: Optional[Any], train: Optional[pd.DataFrame],
                 test: pd.DataFrame, target_column: str, task_type: str,
                 serializer: Optional[Any] = None):
        """Creates an RAIBaseInsights object.

        :param model: The model to compute RAI insights for.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param train: The training dataset including the label column.
            This parameter is optional as some extending downstream classes
            may not require it.
        :type train: pandas.DataFrame
        :param test: The test dataset including the label column.
            Note this parameter is always required.
        :type test: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param task_type: The task to run.
        :type task_type: str
        :param classes: The class labels in the training dataset
        :type classes: numpy.ndarray
        :param serializer: Picklable custom serializer with save and load
            methods for custom model serialization.
            The save method writes the model to file given a parent directory.
            The load method returns the deserialized model from the same
            parent directory.
        :type serializer: object
        """
        self.model = model
        self.train = train
        self.test = test
        self.target_column = target_column
        self.task_type = task_type
        self._serializer = serializer

    @abstractmethod
    def _initialize_managers(self):
        """Initializes the managers.

        This method is abstract and should not be called.
        """
        pass

    @abstractmethod
    def _validate_rai_insights_input_parameters(self, *args):
        """Abstract method to validate the inputs for the constructor."""
        pass

    def compute(self):
        """Calls compute on each of the managers."""
        for manager in self._managers:
            manager.compute()

    def list(self):
        """List information about each of the managers.
        :return: Information about each of the managers.
        :rtype: dict
        """
        configs = {}
        for manager in self._managers:
            configs[manager.name] = manager.list()
        return configs

    def get(self):
        """List information about each of the managers.

        :return: Information about each of the managers.
        :rtype: dict
        """
        configs = {}
        for manager in self._managers:
            configs[manager.name] = manager.get()
        return configs

    @abstractmethod
    def get_data(self):
        """Get all data from RAIBaseInsights object."""
        pass

    @abstractmethod
    def _get_dataset(self):
        pass

    def _write_to_file(self, file_path, content):
        """Save the string content to the given file path.
        :param file_path: The file path to save the content to.
        :type file_path: str
        :param content: The string content to save.
        :type content: str
        """
        with open(file_path, 'w') as file:
            file.write(content)

    @abstractmethod
    def _save_predictions(self, path):
        """Save the predict() and predict_proba() output.

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        pass

    def _save_data(self, path):
        """Save the copy of raw data (train and test sets) and
           their related metadata.

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        data_directory = Path(path) / SerializationAttributes.DATA_DIRECTORY
        data_directory.mkdir(parents=True, exist_ok=True)
        if self.train is not None:
            dtypes = self.train.dtypes.astype(str).to_dict()
            self._write_to_file(data_directory /
                                (Metadata.TRAIN + _DTYPES + FileFormats.JSON),
                                json.dumps(dtypes))
            self._write_to_file(data_directory /
                                (Metadata.TRAIN + FileFormats.JSON),
                                self.train.to_json(orient='split'))

        dtypes = self.test.dtypes.astype(str).to_dict()
        self._write_to_file(data_directory /
                            (Metadata.TEST + _DTYPES + FileFormats.JSON),
                            json.dumps(dtypes))
        self._write_to_file(data_directory /
                            (Metadata.TEST + FileFormats.JSON),
                            self.test.to_json(orient='split'))

        self._write_to_file(Path(path) /
                            (SerializationAttributes.RAI_VERSION_JSON),
                            json.dumps(
                                {"responsibleai": responsibleai.__version__}))

    @abstractmethod
    def _save_metadata(self, path):
        """Save the metadata like target column, categorical features,
           task type and the classes (if any).

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        pass

    def _save_model(self, path):
        """Save the model and the serializer (if any).

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        top_dir = Path(path)
        if self._serializer is not None:
            # save the model
            self._serializer.save(self.model, top_dir)
            # save the serializer
            with open(top_dir / _SERIALIZER, 'wb') as file:
                pickle.dump(self._serializer, file)
        else:
            if self.model is not None:
                has_setstate = hasattr(self.model, '__setstate__')
                has_getstate = hasattr(self.model, '__getstate__')
                if not (has_setstate and has_getstate):
                    warnings.warn(
                        "Model does not have __setstate__ and " +
                        "__getstate__, pickle may fail")
            try:
                with open(top_dir / _MODEL_PKL, 'wb') as file:
                    pickle.dump(self.model, file)
            except Exception as e:
                raise UserConfigValidationException(
                    "Model must be picklable or a custom serializer must"
                    " be specified") from e

    def _save_managers(self, path):
        """Save the state of individual managers.

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        top_dir = Path(path)
        # save each of the individual managers
        for manager in self._managers:
            manager._save(top_dir / manager.name)

    def save(self, path):
        """Save the RAIBaseInsights to the given path.

        :param path: The directory path to save the RAIBaseInsights to.
        :type path: str
        """
        self._save_managers(path)
        self._save_data(path)
        self._save_metadata(path)
        self._save_model(path)
        self._save_predictions(path)

    @staticmethod
    def _load_data(inst, path):
        """Load the raw data (train and test sets).

        :param inst: RAIBaseInsights object instance.
        :type inst: RAIBaseInsights
        :param path: The directory path to data location.
        :type path: str
        """
        data_directory = Path(path) / SerializationAttributes.DATA_DIRECTORY
        train_data_json = data_directory / (Metadata.TRAIN + FileFormats.JSON)
        if train_data_json.exists():
            train_dtypes_path = (Metadata.TRAIN + _DTYPES + FileFormats.JSON)
            train_dtypes_json = data_directory / train_dtypes_path
            with open(train_dtypes_json, 'r') as file:
                types = json.load(file)
            with open(train_data_json, 'r') as file:
                train = pd.read_json(file, dtype=types, orient='split')
        else:
            train = None
        inst.__dict__[Metadata.TRAIN] = train
        with open(data_directory /
                  (Metadata.TEST + _DTYPES + FileFormats.JSON), 'r') as file:
            types = json.load(file)
        with open(data_directory / (Metadata.TEST + FileFormats.JSON),
                  'r') as file:
            test = pd.read_json(file, dtype=types, orient='split')
        inst.__dict__[Metadata.TEST] = test

    @staticmethod
    def _load_model(inst, path):
        """Load the model.

        :param inst: RAIBaseInsights object instance.
        :type inst: RAIBaseInsights
        :param path: The directory path to model location.
        :type path: str
        """
        top_dir = Path(path)
        serializer_path = top_dir / _SERIALIZER
        model_load_err = ('ERROR-LOADING-USER-MODEL: '
                          'There was an error loading the user model.')
        if serializer_path.exists():
            try:
                with open(serializer_path, 'rb') as file:
                    serializer = pickle.load(file)
                inst.__dict__['_' + _SERIALIZER] = serializer
                inst.__dict__[Metadata.MODEL] = serializer.load(top_dir)
            except Exception as e:
                warnings.warn(model_load_err)
                raise e
        else:
            inst.__dict__['_' + _SERIALIZER] = None
            try:
                with open(top_dir / _MODEL_PKL, 'rb') as file:
                    inst.__dict__[Metadata.MODEL] = pickle.load(file)
            except Exception as e:
                warnings.warn(model_load_err)
                raise e

    @staticmethod
    def _load_managers(inst, path, manager_map):
        """Load the specified managers from the given path.

        :param inst: RAIBaseInsights object instance.
        :type inst: RAIBaseInsights
        :param path: The directory path to the location of
            the serialized managers.
        :type path: str
        :param manager_map: The map of manager names to manager classes.
        :type manager_map: dict
        """
        top_dir = Path(path)
        managers = []
        for manager_name, manager_class in manager_map.items():
            full_name = f'_{manager_name}_manager'
            manager_dir = top_dir / manager_name
            manager = manager_class._load(manager_dir, inst)
            inst.__dict__[full_name] = manager
            managers.append(manager)

        inst.__dict__['_' + _MANAGERS] = managers

    @staticmethod
    def _load(path, inst, manager_map, load_metadata_func):
        """Load the RAIBaseInsights from the given path.

        :param path: The directory path to load the RAIBaseInsights from.
        :type path: str
        :param inst: RAIBaseInsights object instance.
        :type inst: RAIBaseInsights
        :param manager_map: The map of manager names to manager classes.
        :type manager_map: dict
        :param load_metadata_func: The function to load the metadata.
        :type load_metadata_func: function
        :return: The RAIBaseInsights object after loading.
        :rtype: RAIBaseInsights
        """
        # load current state
        RAIBaseInsights._load_data(inst, path)
        load_metadata_func(inst, path)
        RAIBaseInsights._load_model(inst, path)
        RAIBaseInsights._load_managers(inst, path, manager_map)

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest
from tests.common_utils import create_adult_income_dataset

# Constants representing columns in the synthetic_data fixture
SYNTHETIC_DATA_LABEL = "Label"
SYNTHETIC_DATA_GENDER = "Gender"
SYNTHETIC_DATA_ETHNICITY = "Ethnicity"


@pytest.fixture(scope="session")
def adult_data():
    (
        data_train,
        data_test,
        y_train,
        y_test,
        categorical_features,
        _,
        target_col,
        classes,
        _,
        _,
    ) = create_adult_income_dataset(create_small_dataset=False)
    data_train[target_col] = y_train
    data_test[target_col] = y_test
    return data_train, data_test, categorical_features, target_col, classes


@pytest.fixture(scope="session")
def adult_data_feature_balance_measures():
    return {
        "1": pd.DataFrame(
            {
                "ClassA": {"0": "White", "1": "Male"},
                "ClassB": {"0": "Other", "1": "Female"},
                "FeatureName": {"0": "race", "1": "gender"},
                "StatisticalParity": {"0": 0.0982367846, "1": 0.1977193125},
                "PointwiseMutualInfo": {"0": 0.4906283898, "1": 1.0465727536},
                "SorensonDiceCoeff": {"0": 0.1394139444, "1": 0.1622153797},
                "JaccardIndex": {"0": 0.1846417746, "1": 0.2230209651},
                "KendallRankCorrelation": {
                    "0": -4.6080688829,
                    "1": -0.8738917288,
                },
                "LogLikelihoodRatio": {"0": 2.264674102, "1": 1.7462342396},
                "TTest": {"0": -1.0210127581, "1": -0.3685341442},
                "TTestPValue": {"0": 0.2466906059, "1": 0.3876079751},
            }
        )
    }


@pytest.fixture(scope="session")
def adult_data_distribution_balance_measures():
    return pd.DataFrame(
        {
            "FeatureName": {"0": "race", "1": "gender"},
            "KLDivergence": {"0": 0.2791392127, "1": 0.0576400449},
            "JensenShannonDist": {"0": 0.2741838602, "1": 0.1209204479},
            "WassersteinDist": {"0": 0.3549600737, "1": 0.168112715},
            "InfiniteNormDist": {"0": 0.3549600737, "1": 0.168112715},
            "TotalVarianceDist": {"0": 0.3549600737, "1": 0.168112715},
            "ChiSquarePValue": {"0": 0.0, "1": 0.0},
            "ChiSquareStat": {"0": 13127.8433660934, "1": 2944.6623157248},
            "CrossEntropy": {"0": 1.5057745222, "1": 1.0865356576},
        }
    )


@pytest.fixture(scope="session")
def adult_data_aggregate_balance_measures():
    return pd.DataFrame(
        {
            "TheilLIndex": {"0": 0.3912215368},
            "TheilTIndex": {"0": 0.3420484048},
            "AtkinsonIndex": {"0": 0.3237696704},
        }
    )


# This synthetic dataset was manually created in the SynapseML data balance
# feature so that metrics can be manually computed, thought of as the expected
# results, and then compared to the actual results. This and the expected
# measure fixtures below it are ports from the open source SynapseML repo.
@pytest.fixture(scope="session")
def synthetic_data():
    rows = [
        (0, "Male", "Asian"),
        (0, "Male", "White"),
        (1, "Male", "Other"),
        (1, "Male", "Black"),
        (0, "Female", "White"),
        (0, "Female", "Black"),
        (1, "Female", "Black"),
        (0, "Other", "Asian"),
        (0, "Other", "White"),
    ]
    return pd.DataFrame(
        rows,
        columns=[
            SYNTHETIC_DATA_LABEL,
            SYNTHETIC_DATA_GENDER,
            SYNTHETIC_DATA_ETHNICITY,
        ],
    )


@pytest.fixture(scope="session")
def expected_aggregate_measures_gender():
    return {
        "AtkinsonIndex": 0.03850028646172776,
        "TheilLIndex": 0.039261011885461196,
        "TheilTIndex": 0.03775534151008828,
    }


@pytest.fixture(scope="session")
def expected_aggregate_measures_gender_ethnicity():
    return {
        "AtkinsonIndex": 0.030659793186437745,
        "TheilLIndex": 0.03113963808639034,
        "TheilTIndex": 0.03624967113471546,
    }


@pytest.fixture(scope="session")
def expected_distribution_measures_gender():
    return {
        "FeatureName": "Gender",
        "KLDivergence": 0.03775534151008829,
        "JensenShannonDist": 0.09785224086736323,
        "WassersteinDist": 0.07407407407407407,
        "InfiniteNormDist": 0.1111111111111111,
        "TotalVarianceDist": 0.1111111111111111,
        "ChiSquarePValue": 0.7165313105737893,
        "ChiSquareStat": 0.6666666666666666,
        "CrossEntropy": 1.6416041676697,
    }


@pytest.fixture(scope="session")
def expected_distribution_measures_ethnicity():
    return {
        "FeatureName": "Ethnicity",
        "KLDivergence": 0.07551068302017659,
        "JensenShannonDist": 0.14172745151398888,
        "WassersteinDist": 0.08333333333333333,
        "InfiniteNormDist": 0.1388888888888889,
        "TotalVarianceDist": 0.16666666666666666,
        "ChiSquarePValue": 0.7476795872877147,
        "ChiSquareStat": 1.222222222222222,
        "CrossEntropy": 2.1274437525244,
    }


@pytest.fixture(scope="session")
def expected_feature_balance_measures_gender():
    # Note: Compared to SynapseML, this module does not compute n_pmi_y,
    # n_pmi_xy, and s_pmi. However, this module does compute TTest and
    # TTestPValue so their actual values are used.
    return {
        "ClassA": "Male",
        "ClassB": "Female",
        "FeatureName": "Gender",
        "StatisticalParity": 0.16666666666666669,
        "PointwiseMutualInfo": 0.4054651081081645,
        "SorensonDiceCoeff": 0.1190476190476191,
        "JaccardIndex": 0.20000000000000004,
        "KendallRankCorrelation": 0.18801108758923135,
        "LogLikelihoodRatio": 0.6931471805599454,
        "TTest": -0.1855414423,
        "TTestPValue": 0.4349585786,
    }


@pytest.fixture(scope="session")
def singular_valued_col_data():
    return pd.DataFrame(
        {
            # Note that the col only has 1 unique value
            "col1": [1, 1, 1, 1, 1],
            "col2": [1, 2, 3, 4, 5],
            "target": [0, 1, 0, 1, 0],
        }
    )



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Data Balance Manager class."""

import json
import warnings
from pathlib import Path
from typing import Dict, List

import numpy as np
import pandas as pd

from responsibleai._interfaces import TaskType
from responsibleai._internal.constants import DataBalanceManagerKeys as Keys
from responsibleai._internal.constants import (FileFormats, ListProperties,
                                               ManagerNames)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.databalanceanalysis import (AggregateBalanceMeasures,
                                               DistributionBalanceMeasures,
                                               FeatureBalanceMeasures)
from responsibleai.databalanceanalysis.data_balance_utils import (
    prepare_df, transform_measures_to_dict)
from responsibleai.managers.base_manager import BaseManager

DATA_JSON = f"data{FileFormats.JSON}"
MANAGER_JSON = f"manager{FileFormats.JSON}"
MEASURES_JSON = f"measures{FileFormats.JSON}"


class DataBalanceManager(BaseManager):
    """
    Defines the DataBalanceManager for finding data balance measures
    on a dataset.
    """

    def __init__(
        self,
        train: pd.DataFrame,
        test: pd.DataFrame,
        target_column: str,
        classes: np.ndarray,
        task_type: TaskType,
    ):
        """
        Creates a DataBalanceManager object.

        :param train: The training data.
        :type train: pd.DataFrame
        :param test: The test data.
        :type test: pd.DataFrame
        :param target_column: The name of the target column.
        :type target_column: str
        :param classes: The classes of the target column.
        :type classes: np.ndarray
        :param task_type: The task type of the data.
        :type task_type: TaskType
        """
        self._train = train
        self._test = test
        self._target_column = target_column
        # Eagerly convert numpy array to list of strs for serialization
        self._classes = (
            list(map(str, classes)) if classes is not None else None
        )
        self._task_type = task_type
        self._is_added = False

        self._df = None
        if self._train is not None or self._test is not None:
            self._df = pd.concat([self._train, self._test])

        # Populated in add()
        self._cols_of_interest = None

        # Populated in compute()
        self._data_balance_measures = None

    def add(self, cols_of_interest: List[str]):
        """
        Add data balance measures to be computed later.

        :param cols_of_interest: The names of the columns to be used
            for computing data balance measures.
        :type cols_of_interest: List[str]
        """
        self._cols_of_interest = cols_of_interest

        # Let user see exceptions early in add() before calling compute()
        self._validate()
        self._is_added = True

    def _validate(self):
        """
        Validate that data balance measures can be computed. Raises ValueError
        if not valid.
        """
        if self._train is None and self._test is None:
            raise ValueError(
                (
                    "Either `train` or `test` must be provided to RAIInsights"
                    " to compute data balance measures."
                )
            )

        if not self._target_column:
            raise ValueError(
                (
                    "The `target_column` must be provided to compute"
                    " data balance measures."
                )
            )

        if self._target_column not in self._df.columns:
            raise ValueError(
                (
                    f"The target_column '{self._target_column}' must be"
                    " present in the dataset."
                )
            )

        if not self._classes:
            raise ValueError(
                (
                    f"The target_column '{self._target_column}' must have"
                    " at least one value."
                )
            )

        if self._task_type != TaskType.CLASSIFICATION:
            raise ValueError(
                (
                    f"The task_type '{self._task_type}' must be"
                    " 'classification' to compute data balance measures."
                )
            )

        if not self._cols_of_interest:
            raise ValueError(
                (
                    "The `cols_of_interest` must be provided to compute"
                    " data balance measures."
                )
            )

        if not all(col in self._df.columns for col in self._cols_of_interest):
            raise ValueError(
                (
                    "All columns in `cols_of_interest` must be present in"
                    " the dataset."
                )
            )

    def compute(self):
        """
        Computes data balance measures on the dataset.
        """
        if not self._is_added:
            return

        try:
            self._validate()

            self._df = prepare_df(df=self._df)

            feature_balance_measures = {}
            for pos_label in self._classes:
                feature_balance_measures[pos_label] = FeatureBalanceMeasures(
                    cols_of_interest=self._cols_of_interest,
                    label_col=self._target_column,
                    pos_label=pos_label,
                ).measures(dataset=self._df)

            distribution_balance_measures = DistributionBalanceMeasures(
                cols_of_interest=self._cols_of_interest
            ).measures(dataset=self._df)
            aggregate_balance_measures = AggregateBalanceMeasures(
                cols_of_interest=self._cols_of_interest
            ).measures(dataset=self._df)

            self._set_data_balance_measures(
                feature_balance_measures=feature_balance_measures,
                distribution_balance_measures=distribution_balance_measures,
                aggregate_balance_measures=aggregate_balance_measures,
            )
        except Exception as e:
            warnings.warn(
                f"Failed to compute data balance measures due to {e!r}."
            )

    def _set_data_balance_measures(
        self,
        feature_balance_measures: Dict[str, pd.DataFrame],
        distribution_balance_measures: pd.DataFrame,
        aggregate_balance_measures: pd.DataFrame,
    ):
        """
        Take the data balance measure DataFrames, transform them into
        a dictionary, and set them as an instance variable.

        :param feature_balance_measures: Feature balance measures.
        :type feature_balance_measures: Dict[str, pd.DataFrame]
        :param distribution_balance_measures: Distribution balance measures.
        :type distribution_balance_measures: pd.DataFrame
        :param aggregate_balance_measures: Aggregate balance measures.
        :type aggregate_balance_measures: pd.DataFrame
        """
        self._data_balance_measures = transform_measures_to_dict(
            feature_balance_measures=feature_balance_measures,
            distribution_balance_measures=distribution_balance_measures,
            aggregate_balance_measures=aggregate_balance_measures,
        )

    def get(self):
        """
        Get the computed data balance measures.

        :return: The computed data balance measures.
        :rtype: Dict[str, Any]
        """
        return (
            self._data_balance_measures if self._data_balance_measures else {}
        )

    def list(self):
        """
        List information about the data balance manager.

        :return: Information about the data balance manager.
        :rtype: Dict[str, Any]
        """
        props = {
            ListProperties.MANAGER_TYPE: self.name,
            Keys.IS_ADDED: self._is_added,
            Keys.TASK_TYPE: self._task_type,
            Keys.COLS_OF_INTEREST: self._cols_of_interest,
            Keys.TARGET_COLUMN: self._target_column,
            Keys.CLASSES: self._classes,
        }

        return props

    def get_data(self):
        """
        Get the computed data balance measures. This is the same as calling
        get() because data balance measures are part of the Dataset interface
        rather than on their own (such as CausalData, CounterfactualData, etc).

        :return: The computed data balance measures.
        :rtype: Dict[str, Any]
        """
        return self.get()

    @property
    def name(self):
        return ManagerNames.DATA_BALANCE

    def _save(self, path):
        """
        Save the DataBalanceManager to the given path.

        :param path: The directory path to save the DataBalanceManager to.
        :type path: str
        """
        top_dir = Path(path)
        top_dir.mkdir(parents=True, exist_ok=True)

        # If manager is not added to RAIInsights, don't save any of its configs
        if not self._is_added:
            return

        dir_manager = DirectoryManager(parent_directory_path=path)
        config_dir = dir_manager.create_config_directory()

        manager_path = config_dir / MANAGER_JSON
        with open(manager_path, "w") as f:
            json.dump(self.list(), f)

        # if measures have been computed, save the computed measures
        if self._data_balance_measures:
            measures_path = config_dir / MEASURES_JSON
            with open(measures_path, "w") as f:
                json.dump(self._data_balance_measures, f)

        data_path = dir_manager.create_data_directory() / DATA_JSON
        self._df.to_json(data_path, orient="split")

    @staticmethod
    def _load(path, rai_insights):
        """
        Load the DataBalanceManager from the given path.

        :param path: The directory path to load the DataBalanceManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The DataBalanceManager after loading.
        :rtype: DataBalanceManager
        """
        # create the DataBalanceManager without any properties using the
        # __new__ function, similar to pickle
        inst = DataBalanceManager.__new__(DataBalanceManager)

        inst.__dict__["_train"] = rai_insights.train
        inst.__dict__["_test"] = rai_insights.test

        is_added = False
        cols_of_interest = None
        task_type = rai_insights.task_type
        target_column = rai_insights.target_column
        classes = (
            list(map(str, rai_insights._classes))
            if rai_insights._classes is not None
            else []
        )
        df = pd.concat([rai_insights.train, rai_insights.test])
        data_balance_measures = None

        all_db_dirs = DirectoryManager.list_sub_directories(path)
        if len(all_db_dirs) != 0:
            dir_manager = DirectoryManager(
                parent_directory_path=path, sub_directory_name=all_db_dirs[0]
            )
            config_dir = dir_manager.get_config_directory()

            # Load manager
            with open(config_dir / MANAGER_JSON, "r") as f:
                manager_info = json.load(f)
                is_added = manager_info[Keys.IS_ADDED]
                task_type = manager_info[Keys.TASK_TYPE]
                cols_of_interest = manager_info[Keys.COLS_OF_INTEREST]
                target_column = manager_info[Keys.TARGET_COLUMN]
                classes = manager_info[Keys.CLASSES]

            # Load from data json
            data_path = dir_manager.get_data_directory() / DATA_JSON
            if data_path.exists():
                df = pd.read_json(data_path, orient="split")

            # Load measures
            measures_path = config_dir / MEASURES_JSON
            if measures_path.exists():
                with open(measures_path, "r") as f:
                    data_balance_measures = json.load(f)

        inst.__dict__["_is_added"] = is_added
        inst.__dict__["_task_type"] = task_type
        inst.__dict__["_cols_of_interest"] = cols_of_interest
        inst.__dict__["_target_column"] = target_column
        inst.__dict__["_classes"] = classes
        inst.__dict__["_df"] = df
        inst.__dict__["_data_balance_measures"] = data_balance_measures

        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the Error Analysis Manager class."""

import json
from pathlib import Path
from typing import Any, List, Optional

import jsonschema
import pandas as pd

from erroranalysis._internal.constants import metric_to_display_name
from erroranalysis._internal.error_analyzer import ModelAnalyzer
from erroranalysis._internal.error_report import as_error_report
from erroranalysis._internal.error_report import \
    json_converter as report_json_converter
from raiutils.exceptions import UserConfigValidationException
from responsibleai._config.base_config import BaseConfig
from responsibleai._interfaces import ErrorAnalysisData
from responsibleai._internal.constants import ErrorAnalysisManagerKeys as Keys
from responsibleai._internal.constants import (FileFormats, ListProperties,
                                               ManagerNames)
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager
from responsibleai.exceptions import (ConfigAndResultMismatchException,
                                      DuplicateManagerConfigException)
from responsibleai.managers.base_manager import BaseManager
from responsibleai.utils import _measure_time

REPORTS = 'reports'
CONFIG = 'config'
MAX_DEPTH = Keys.MAX_DEPTH
NUM_LEAVES = Keys.NUM_LEAVES
MIN_CHILD_SAMPLES = Keys.MIN_CHILD_SAMPLES
FILTER_FEATURES = Keys.FILTER_FEATURES
IS_COMPUTED = 'is_computed'


def config_json_converter(obj):
    """Helper function to convert ErrorAnalysisConfig object to json.

    :param obj: Object to convert to json.
    :type obj: object
    :return: The converted json.
    :rtype: dict
    """
    if isinstance(obj, ErrorAnalysisConfig):
        return obj.__dict__
    try:
        return obj.to_json()
    except AttributeError:
        return obj.__dict__


def as_error_config(json_dict):
    """Helper function to convert json to an ErrorAnalysisConfig object.

    :param json_dict: The json to convert.
    :type json_dict: dict
    :return: The converted ErrorAnalysisConfig.
    :rtype: ErrorAnalysisConfig
    """
    has_max_depth = MAX_DEPTH in json_dict
    has_num_leaves = NUM_LEAVES in json_dict
    has_min_child_samples = MIN_CHILD_SAMPLES in json_dict
    has_filter_features = FILTER_FEATURES in json_dict
    has_is_computed = IS_COMPUTED in json_dict
    has_all_fields = (has_max_depth and has_num_leaves and
                      has_min_child_samples and has_filter_features and
                      has_is_computed)
    if has_all_fields:
        max_depth = json_dict[MAX_DEPTH]
        num_leaves = json_dict[NUM_LEAVES]
        min_child_samples = json_dict[MIN_CHILD_SAMPLES]
        filter_features = json_dict[FILTER_FEATURES]
        config = ErrorAnalysisConfig(max_depth,
                                     num_leaves,
                                     min_child_samples,
                                     filter_features)
        config.is_computed = json_dict[IS_COMPUTED]
        return config
    else:
        return json_dict


def get_wrapped_model(model, dropped_features):
    predict_proba_flag = hasattr(model, 'predict_proba')
    if predict_proba_flag:
        wrapper_model = MetadataRemovalClassificationModelWrapper(
            model,
            dropped_features)
    else:
        wrapper_model = MetadataRemovalRegressionModelWrapper(
            model,
            dropped_features)
    return wrapper_model


class MetadataRemovalClassificationModelWrapper():
    """Defines MetadataRemovalClassificationModelWrapper, wrapping the
    classification model to ignore dropped feature metadata if any."""

    def __init__(self, model: any,
                 dropped_features: Optional[List[str]] = None):
        """If needed, wraps classification model to ignore dropped features.

        :param model: The model or function to evaluate on the examples.
        :type model: function or model with a predict or predict_proba function
        :param dropped_features: List of features that were dropped by the
                                 the user during training of their model.
        :type dropped_features: Optional[List[str]]
        """
        self.model = model
        self.dropped_features = dropped_features

    def predict(self, dataset: pd.DataFrame):
        return self._apply_func(self.model.predict, dataset)

    def predict_proba(self, dataset: pd.DataFrame):
        return self._apply_func(self.model.predict_proba, dataset)

    def _apply_func(self, func, dataset):
        if self.dropped_features is None or len(self.dropped_features) == 0:
            return func(dataset)
        return func(dataset.drop(columns=self.dropped_features, axis=1))


class MetadataRemovalRegressionModelWrapper():
    """Defines MetadataRemovalRegressionModelWrapper, wrapping the
    regression model to ignore dropped feature metadata if any."""

    def __init__(self, model: any,
                 dropped_features: Optional[List[str]] = None):
        """If needed, wraps the model to ignore the dropped features.

        :param model: The model or function to evaluate on the examples.
        :type model: function or model with a predict function
        :param dropped_features: List of features that were dropped by the
                                 the user during training of their model.
        :type dropped_features: Optional[List[str]]
        """
        self.model = model
        self.dropped_features = dropped_features

    def predict(self, dataset: pd.DataFrame):
        if self.dropped_features is None or len(self.dropped_features) == 0:
            return self.model.predict(dataset)
        return self.model.predict(dataset.drop(
            columns=self.dropped_features, axis=1))


class ErrorAnalysisConfig(BaseConfig):
    """Defines the ErrorAnalysisConfig, specifying the parameters to run."""

    def __init__(self, max_depth: int, num_leaves: int,
                 min_child_samples: int, filter_features: List):
        """Creates an ErrorAnalysisConfig, specifying the parameters to run.

        :param max_depth: The maximum depth of the tree.
        :type max_depth: int
        :param num_leaves: The number of leaves in the tree.
        :type num_leaves: int
        :param min_child_samples: The minimal number of data required to
            create one leaf.
        :type min_child_samples: int
        :param filter_features: One or two features to use for the
            matrix filter.
        :type filter_features: list
        """
        super(ErrorAnalysisConfig, self).__init__()
        self.max_depth = max_depth
        self.num_leaves = num_leaves
        self.min_child_samples = min_child_samples
        self.filter_features = filter_features

    def __eq__(self, other_ea_config):
        """Returns true if this config is equal to the given config.

        :return: True if the given config is equal to the current config.
        :rtype: boolean
        """
        return (
            self.max_depth == other_ea_config.max_depth and
            self.num_leaves == other_ea_config.num_leaves and
            self.min_child_samples == other_ea_config.min_child_samples and
            self.filter_features == other_ea_config.filter_features
        )

    @property
    def __dict__(self):
        """Returns the dictionary representation of the ErrorAnalysisConfig.

        The dictionary contains the max depth, num leaves, min
        child samples and list of matrix filter features.

        :return: The dictionary representation of the ErrorAnalysisConfig.
        :rtype: dict
        """
        return {'max_depth': self.max_depth,
                'num_leaves': self.num_leaves,
                'min_child_samples': self.min_child_samples,
                'filter_features': self.filter_features,
                'is_computed': self.is_computed}

    def to_json(self):
        """Serialize ErrorAnalysisConfig object to json.

        :return: The string json representation of the ErrorAnalysisConfig.
        :rtype: str
        """
        return json.dumps(self, default=config_json_converter, indent=2)

    @staticmethod
    def from_json(json_str):
        """Deserialize json string to an ErrorAnalysisConfig object.

        :return: The deserialized ErrorAnalysisConfig.
        :rtype: ErrorAnalysisConfig
        """
        return json.loads(json_str, object_hook=as_error_config)


class ErrorAnalysisManager(BaseManager):
    """Defines the ErrorAnalysisManager for discovering errors in a model."""

    def __init__(self, model: Any, dataset: pd.DataFrame, target_column: str,
                 classes: Optional[List] = None,
                 categorical_features: Optional[List[str]] = None,
                 dropped_features: Optional[List[str]] = None):
        """Creates an ErrorAnalysisManager object.

        :param model: The model to analyze errors on.
            A model that implements sklearn.predict or sklearn.predict_proba
            or function that accepts a 2d ndarray.
        :type model: object
        :param dataset: The dataset including the label column.
        :type dataset: pandas.DataFrame
        :param target_column: The name of the label column.
        :type target_column: str
        :param classes: Class names as a list of strings.
            The order of the class names should match that of the model
            output.  Only required if analyzing a classifier.
        :type classes: list
        :param categorical_features: The categorical feature names.
        :type categorical_features: list[str]
        :param dropped_features: List of features that are omitted for model
                            training. This includes metadata that is useful
                            for evaluating the model.
        :type dropped_features: Optional[List[str]]
        """
        self._true_y = dataset[target_column]
        self._dataset = dataset.drop(columns=[target_column])
        self._feature_names = list(self._dataset.columns)
        self._classes = classes
        self._categorical_features = categorical_features
        self._ea_config_list = []
        self._ea_report_list = []
        wrapper_model = get_wrapped_model(
            model,
            dropped_features)
        self._analyzer = ModelAnalyzer(
            wrapper_model,
            self._dataset,
            self._true_y,
            self._feature_names,
            self._categorical_features,
            classes=self._classes)

    def add(self, max_depth: int = 3, num_leaves: int = 31,
            min_child_samples: int = 20,
            filter_features: Optional[List] = None):
        """Add an error analyzer to be computed later.

        :param max_depth: The maximum depth of the tree.
        :type max_depth: Optional[int]
        :param num_leaves: The number of leaves in the tree.
        :type num_leaves: Optional[int]
        :param min_child_samples: The minimal number of data required to
            create one leaf.
        :type min_child_samples: Optional[int]
        :param filter_features: One or two features to use for the
            matrix filter.
        :type filter_features: Optional[list]
        """
        model_wrapper_without_metadata = self._analyzer.model
        if model_wrapper_without_metadata is None or \
                model_wrapper_without_metadata.model is None:
            raise UserConfigValidationException(
                'Model is required for error analysis')

        ea_config = ErrorAnalysisConfig(
            max_depth=max_depth,
            num_leaves=num_leaves,
            min_child_samples=min_child_samples,
            filter_features=filter_features)
        is_duplicate = ea_config.is_duplicate(
            self._ea_config_list)

        if is_duplicate:
            raise DuplicateManagerConfigException(
                "Duplicate config specified for error analysis,"
                "config already added")
        else:
            self._ea_config_list.append(ea_config)

    @_measure_time
    def compute(self):
        """Creates an ErrorReport by running the error analyzer on the model.
        """
        print("Error Analysis")
        print('Current Status: Generating error analysis reports.')
        for config in self._ea_config_list:
            if config.is_computed:
                continue
            config.is_computed = True
            max_depth = config.max_depth
            num_leaves = config.num_leaves
            min_child_samples = config.min_child_samples
            filter_features = config.filter_features
            report = self._analyzer.create_error_report(
                filter_features, max_depth=max_depth,
                min_child_samples=min_child_samples,
                num_leaves=num_leaves,
                compute_importances=True,
                compute_root_stats=True)

            # Validate the serialized output against schema
            schema = ErrorAnalysisManager._get_error_analysis_schema()
            jsonschema.validate(
                json.loads(report.to_json()), schema)

            self._ea_report_list.append(report)
        print('Current Status: Finished generating error analysis reports.')

    def get(self):
        """Get the computed error reports.

        Must be called after add and compute methods.

        :return: The computed error reports.
        :rtype: list[erroranalysis._internal.error_report.ErrorReport]
        """
        return self._ea_report_list

    @staticmethod
    def _get_error_analysis_schema():
        """Get the schema for validating the error analysis output."""
        schema_directory = (Path(__file__).parent.parent / '_tools' /
                            'error_analysis' / 'dashboard_schemas')
        schema_filename = f'error_analysis_output_v0.0{FileFormats.JSON}'
        schema_filepath = schema_directory / schema_filename
        with open(schema_filepath, 'r') as f:
            return json.load(f)

    def list(self):
        """List information about the ErrorAnalysisManager.

        :return: A dictionary of properties.
        :rtype: dict
        """
        props = {ListProperties.MANAGER_TYPE: self.name}
        reports = []
        for config in self._ea_config_list:
            report = {}
            report[Keys.IS_COMPUTED] = config.is_computed
            report[Keys.MAX_DEPTH] = config.max_depth
            report[Keys.NUM_LEAVES] = config.num_leaves
            report[Keys.MIN_CHILD_SAMPLES] = config.min_child_samples
            report[Keys.FILTER_FEATURES] = config.filter_features
            reports.append(report)
        props[Keys.REPORTS] = reports
        return props

    def get_data(self):
        """Get error analysis data

        :return: A array of ErrorAnalysisConfig.
        :rtype: List[ErrorAnalysisConfig]
        """
        report_props = zip(self.get(), self.list()[Keys.REPORTS])
        return [
            self._get_error_analysis(report,
                                     props) for report, props in report_props]

    def _get_error_analysis(self, report, props):
        error_analysis = ErrorAnalysisData()
        error_analysis.maxDepth = props[Keys.MAX_DEPTH]
        error_analysis.numLeaves = props[Keys.NUM_LEAVES]
        error_analysis.minChildSamples = props[Keys.MIN_CHILD_SAMPLES]
        error_analysis.tree = report.tree
        error_analysis.matrix = report.matrix
        error_analysis.importances = report.importances
        error_analysis.metric = metric_to_display_name[self._analyzer.metric]
        error_analysis.root_stats = report.root_stats
        return error_analysis

    @property
    def name(self):
        """Get the name of the error analysis manager.

        :return: The name of the error analysis manager.
        :rtype: str
        """
        return ManagerNames.ERROR_ANALYSIS

    def _save(self, path):
        """Save the ErrorAnalysisManager to the given path.

        :param path: The directory path to save the ErrorAnalysisManager to.
        :type path: str
        """
        top_dir = Path(path)
        top_dir.mkdir(parents=True, exist_ok=True)

        if len(self._ea_config_list) != len(self._ea_report_list):
            raise ConfigAndResultMismatchException(
                "The number of error analysis configs {0} doesn't match the "
                "number of results {1}".format(
                    len(self._ea_config_list),
                    len(self._ea_report_list)
                )
            )

        for index in range(0, len(self._ea_report_list)):
            # save the configs
            directory_manager = DirectoryManager(parent_directory_path=path)
            config_path = (directory_manager.create_config_directory() /
                           f'config{FileFormats.JSON}')
            ea_config = self._ea_config_list[index]
            with open(config_path, 'w') as file:
                json.dump(ea_config, file,
                          default=config_json_converter)

            # save the reports
            report_path = (directory_manager.create_data_directory() /
                           f'report{FileFormats.JSON}')
            ea_report = self._ea_report_list[index]
            with open(report_path, 'w') as file:
                json.dump(ea_report, file,
                          default=report_json_converter)

    @staticmethod
    def _load(path, rai_insights):
        """Load the ErrorAnalysisManager from the given path.

        :param path: The directory path to load the ErrorAnalysisManager from.
        :type path: str
        :param rai_insights: The loaded parent RAIInsights.
        :type rai_insights: RAIInsights
        :return: The ErrorAnalysisManager manager after loading.
        :rtype: ErrorAnalysisManager
        """
        # create the ErrorAnalysisManager without any properties using
        # the __new__ function, similar to pickle
        inst = ErrorAnalysisManager.__new__(ErrorAnalysisManager)

        ea_config_list = []
        ea_report_list = []
        all_ea_dirs = DirectoryManager.list_sub_directories(path)
        for ea_dir in all_ea_dirs:
            directory_manager = DirectoryManager(
                parent_directory_path=path,
                sub_directory_name=ea_dir)

            config_path = (directory_manager.get_config_directory() /
                           f'config{FileFormats.JSON}')
            with open(config_path, 'r') as file:
                ea_config = json.load(file, object_hook=as_error_config)
                ea_config_list.append(ea_config)

            report_path = (directory_manager.get_data_directory() /
                           f'report{FileFormats.JSON}')
            with open(report_path, 'r') as file:
                ea_report = json.load(file, object_hook=as_error_report)
                # Validate the serialized output against schema
                schema = ErrorAnalysisManager._get_error_analysis_schema()
                jsonschema.validate(
                    json.loads(ea_report.to_json()), schema)
                ea_report_list.append(ea_report)

        inst.__dict__['_ea_report_list'] = ea_report_list
        inst.__dict__['_ea_config_list'] = ea_config_list

        categorical_features = rai_insights.categorical_features
        inst.__dict__['_categorical_features'] = categorical_features
        target_column = rai_insights.target_column
        true_y = rai_insights.test[target_column]
        dataset = rai_insights.test.drop(columns=[target_column])
        inst.__dict__['_dataset'] = dataset
        inst.__dict__['_true_y'] = true_y
        feature_names = list(dataset.columns)
        inst.__dict__['_feature_names'] = feature_names
        dropped_features = None
        if rai_insights._feature_metadata is not None:
            dropped_features = rai_insights._feature_metadata.dropped_features
        inst.__dict__['_dropped_features'] = dropped_features
        wrapper_model = get_wrapped_model(
            rai_insights.model,
            dropped_features)
        inst.__dict__['_analyzer'] = ModelAnalyzer(
            wrapper_model,
            dataset,
            true_y,
            feature_names,
            categorical_features)
        return inst



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd
import pytest

from responsibleai.databalanceanalysis.constants import Constants
from responsibleai.databalanceanalysis.data_balance_utils import (
    AGGREGATE_BALANCE_MEASURES_KEY, DISTRIBUTION_BALANCE_MEASURES_KEY,
    FEATURE_BALANCE_MEASURES_KEY, prepare_df,
    transform_aggregate_balance_measures,
    transform_distribution_balance_measures,
    transform_feature_balance_measures, transform_measures_to_dict)


class TestDataBalanceUtils:
    def test_prepare_df_ensure_cloned(self, adult_data):
        train_df, _, _, _, _ = adult_data
        output = prepare_df(df=train_df)
        assert train_df.equals(output)  # Ensure both dfs have the same data
        assert train_df is not output  # but are different objects

    def test_prepare_df_with_null_input(self):
        with pytest.raises(ValueError):
            output = prepare_df(df=None)
            assert output is None

    def test_prepare_df_with_empty_df(self):
        empty_df = pd.DataFrame()
        output = prepare_df(df=empty_df)
        # Should return an empty df instead of erroring
        assert output.empty
        assert empty_df.equals(output)
        assert empty_df is not output

    def test_transform_measures_to_dict_with_valid_input(
        self,
        adult_data_feature_balance_measures,
        adult_data_distribution_balance_measures,
        adult_data_aggregate_balance_measures,
    ):
        d = transform_measures_to_dict(
            adult_data_feature_balance_measures,
            adult_data_distribution_balance_measures,
            adult_data_aggregate_balance_measures,
        )
        assert isinstance(d, dict)
        assert len(d) == 3

        assert FEATURE_BALANCE_MEASURES_KEY in d
        assert isinstance(d[FEATURE_BALANCE_MEASURES_KEY], dict)
        assert len(d[FEATURE_BALANCE_MEASURES_KEY]) != 0

        assert DISTRIBUTION_BALANCE_MEASURES_KEY in d
        assert isinstance(d[DISTRIBUTION_BALANCE_MEASURES_KEY], dict)
        assert len(d[DISTRIBUTION_BALANCE_MEASURES_KEY]) != 0

        assert AGGREGATE_BALANCE_MEASURES_KEY in d
        assert isinstance(d[AGGREGATE_BALANCE_MEASURES_KEY], dict)
        assert len(d[AGGREGATE_BALANCE_MEASURES_KEY]) != 0

    @pytest.mark.parametrize("measures", [None, pd.DataFrame()])
    def test_transform_measures_to_dict_with_invalid_input(self, measures):
        d = transform_measures_to_dict(
            feature_balance_measures=measures,
            distribution_balance_measures=measures,
            aggregate_balance_measures=measures,
        )
        # Even with invalid input, should return a dict of empty dicts
        assert isinstance(d, dict)
        assert len(d) == 3

        assert FEATURE_BALANCE_MEASURES_KEY in d
        assert isinstance(d[FEATURE_BALANCE_MEASURES_KEY], dict)
        assert len(d[FEATURE_BALANCE_MEASURES_KEY]) == 0

        assert DISTRIBUTION_BALANCE_MEASURES_KEY in d
        assert isinstance(d[DISTRIBUTION_BALANCE_MEASURES_KEY], dict)
        assert len(d[DISTRIBUTION_BALANCE_MEASURES_KEY]) == 0

        assert AGGREGATE_BALANCE_MEASURES_KEY in d
        assert isinstance(d[AGGREGATE_BALANCE_MEASURES_KEY], dict)
        assert len(d[AGGREGATE_BALANCE_MEASURES_KEY]) == 0

    def test_transform_feature_balance_measures_with_valid_input(
        self, adult_data_feature_balance_measures
    ):
        pos_label = "1"

        d = transform_feature_balance_measures(
            dfs=adult_data_feature_balance_measures
        )
        assert isinstance(d, dict)
        assert len(d) != 0

        # FeatureNames in the df are transformed to keys of the dict
        feature_names = adult_data_feature_balance_measures[pos_label][
            Constants.FEATURE_NAME.value
        ].unique()
        for col in feature_names:
            assert col in d[pos_label]

    @pytest.mark.parametrize("dfs", [None, pd.DataFrame(), {}])
    def test_transform_feature_balance_measures_with_invalid_input(self, dfs):
        d = transform_feature_balance_measures(dfs=dfs)
        assert isinstance(d, dict)
        assert len(d) == 0

    def test_transform_distribution_balance_measures_with_valid_input(
        self, adult_data_distribution_balance_measures
    ):
        d = transform_distribution_balance_measures(
            df=adult_data_distribution_balance_measures
        )
        assert isinstance(d, dict)
        assert len(d) != 0

        # FeatureNames in the df are transformed to keys of the dict
        feature_names = adult_data_distribution_balance_measures[
            Constants.FEATURE_NAME.value
        ].unique()
        for col in feature_names:
            assert col in d

    @pytest.mark.parametrize("df", [None, pd.DataFrame()])
    def test_transform_distribution_balance_measures_with_invalid_input(
        self, df
    ):
        d = transform_distribution_balance_measures(df=df)
        assert isinstance(d, dict)
        assert len(d) == 0

    def test_transform_aggregate_balance_measures_with_valid_input(
        self, adult_data_aggregate_balance_measures
    ):
        d = transform_aggregate_balance_measures(
            df=adult_data_aggregate_balance_measures
        )
        assert isinstance(d, dict)
        assert len(d) != 0

    @pytest.mark.parametrize("df", [None, pd.DataFrame()])
    def test_transform_aggregate_balance_measures_with_invalid_input(self, df):
        d = transform_aggregate_balance_measures(df=df)
        assert isinstance(d, dict)
        assert len(d) == 0




# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import os

import pytest

from rai_test_utils.models.lightgbm import create_lightgbm_classifier
from raiutils.exceptions import UserConfigValidationException
from responsibleai import RAIInsights
from responsibleai._interfaces import CounterfactualData
from responsibleai._internal.constants import FileFormats
from responsibleai._tools.shared.state_directory_management import \
    DirectoryManager

from ..common_utils import create_iris_data


class TestCounterfactualAdvancedFeatures(object):

    @pytest.mark.parametrize('vary_all_features', [True, False])
    @pytest.mark.parametrize('feature_importance', [True, False])
    def test_counterfactual_vary_features(
            self, vary_all_features, feature_importance):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')

        if vary_all_features:
            features_to_vary = 'all'
        else:
            features_to_vary = [feature_names[0]]

        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=2,
            features_to_vary=features_to_vary,
            feature_importance=feature_importance)
        rai_insights.counterfactual.compute()

        cf_obj = rai_insights.counterfactual.get()[0]
        assert cf_obj is not None

    @pytest.mark.parametrize('feature_importance', [True, False])
    def test_counterfactual_permitted_range(self, feature_importance):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')

        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=2,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]},
            feature_importance=feature_importance)
        rai_insights.counterfactual.compute()

        cf_obj = rai_insights.counterfactual.get()[0]
        assert cf_obj is not None

    def test_counterfactual_manager_save_load(self, tmpdir):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')

        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=2,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]})
        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=1,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]})
        rai_insights.counterfactual.compute()

        counterfactual_config_list_before_save = \
            rai_insights.counterfactual._counterfactual_config_list
        assert len(counterfactual_config_list_before_save) == 2
        assert len(rai_insights.counterfactual.get()) == 2
        cf_obj_1 = rai_insights.counterfactual.get()[0]
        assert cf_obj_1 is not None
        cf_obj_2 = rai_insights.counterfactual.get()[1]
        assert cf_obj_2 is not None

        save_dir = tmpdir.mkdir('save-dir')
        rai_insights.save(save_dir)
        rai_insights_copy = RAIInsights.load(save_dir)

        counterfactual_config_list_after_save = \
            rai_insights_copy.counterfactual._counterfactual_config_list
        assert len(rai_insights_copy.counterfactual.get()) == 2
        cf_obj_1 = rai_insights_copy.counterfactual.get()[0]
        assert cf_obj_1 is not None
        cf_obj_2 = rai_insights_copy.counterfactual.get()[1]
        assert cf_obj_2 is not None

        assert counterfactual_config_list_before_save[0].id in \
            [counterfactual_config_list_after_save[0].id,
             counterfactual_config_list_after_save[1].id]
        assert counterfactual_config_list_before_save[1].id in \
            [counterfactual_config_list_after_save[0].id,
             counterfactual_config_list_after_save[1].id]

        # Delete the dice-ml explainer directory so that the dice-ml
        # explainer can be re-trained rather being loaded from the
        # disc
        counterfactual_path = save_dir / "counterfactual"
        all_cf_dirs = DirectoryManager.list_sub_directories(
            counterfactual_path)
        for counterfactual_config_dir in all_cf_dirs:
            directory_manager = DirectoryManager(
                parent_directory_path=counterfactual_path,
                sub_directory_name=counterfactual_config_dir)
            explainer_pkl_path = (
                directory_manager.get_generators_directory() /
                ("explainer" + FileFormats.PKL))
            os.remove(explainer_pkl_path)

        with pytest.warns(UserWarning,
                          match='ERROR-LOADING-COUNTERFACTUAL-EXPLAINER: '
                                'There was an error loading the '
                                'counterfactual explainer model. '
                                'Retraining the counterfactual '
                                'explainer.'):
            rai_insights_copy_new = RAIInsights.load(save_dir)
        counterfactual_config_list = \
            rai_insights_copy_new.counterfactual._counterfactual_config_list
        assert len(counterfactual_config_list) == 2
        assert counterfactual_config_list[0].explainer is not None
        assert counterfactual_config_list[1].explainer is not None

    @pytest.mark.parametrize('feature_importance', [True, False])
    def test_counterfactual_manager_request_counterfactuals(
            self, feature_importance):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')

        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=2,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]},
            feature_importance=feature_importance)
        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=1,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]},
            feature_importance=feature_importance)
        rai_insights.counterfactual.compute()

        test_instance = X_test.iloc[0:1].drop('target', axis=1)
        query_id = \
            rai_insights.counterfactual._counterfactual_config_list[0].id
        counterfactual_data = \
            rai_insights.counterfactual.request_counterfactuals(
                query_id, test_instance)

        self._verify_counterfactual_data([counterfactual_data])

        test_instance = X_test.iloc[0:2].drop('target', axis=1)
        query_id = \
            rai_insights.counterfactual._counterfactual_config_list[1].id
        with pytest.raises(
            UserConfigValidationException,
                match='Only one row of data is allowed for '
                      'counterfactual generation.'):
            rai_insights.counterfactual.request_counterfactuals(
                query_id, test_instance)

        test_instance = X_test.iloc[0:1].drop('target', axis=1)
        incorrect_query_id = "incorrect_query_id"
        with pytest.raises(
            UserConfigValidationException,
                match='No counterfactual config found for id {0}.'.format(
                    incorrect_query_id)):
            rai_insights.counterfactual.request_counterfactuals(
                incorrect_query_id, test_instance)

        test_instance = X_test.iloc[0:1].drop('target', axis=1).values
        query_id = \
            rai_insights.counterfactual._counterfactual_config_list[1].id
        with pytest.raises(
            UserConfigValidationException,
                match='Data is of type <class \'numpy.ndarray\'>'
                      ' but it must be a pandas DataFrame.'):
            rai_insights.counterfactual.request_counterfactuals(
                query_id, test_instance)

    def _verify_counterfactual_data(self, counterfactual_data_list):
        assert counterfactual_data_list is not None
        assert len(counterfactual_data_list) == 1
        assert isinstance(counterfactual_data_list[0], CounterfactualData)
        assert hasattr(counterfactual_data_list[0], 'id')
        assert hasattr(counterfactual_data_list[0], 'cfs_list')
        assert hasattr(counterfactual_data_list[0], 'desired_class')
        assert hasattr(counterfactual_data_list[0], 'desired_range')
        assert hasattr(counterfactual_data_list[0], 'feature_names')
        assert hasattr(
            counterfactual_data_list[0], 'feature_names_including_target')
        assert hasattr(counterfactual_data_list[0], 'local_importance')
        assert hasattr(counterfactual_data_list[0], 'summary_importance')
        assert hasattr(counterfactual_data_list[0], 'model_type')
        assert hasattr(counterfactual_data_list[0], 'test_data')
        assert len(counterfactual_data_list[
            0].feature_names_including_target) == \
            len(counterfactual_data_list[0].test_data[0][0])
        assert len(counterfactual_data_list[
            0].feature_names_including_target) == \
            len(counterfactual_data_list[0].cfs_list[0][0])

    def test_counterfactual_manager_get_data(self):
        X_train, X_test, y_train, y_test, feature_names, _ = \
            create_iris_data()

        model = create_lightgbm_classifier(X_train, y_train)
        X_train['target'] = y_train
        X_test['target'] = y_test

        rai_insights = RAIInsights(
            model=model,
            train=X_train,
            test=X_test.iloc[0:10],
            target_column='target',
            task_type='classification')

        rai_insights.counterfactual.add(
            total_CFs=10, desired_class=2,
            features_to_vary=[feature_names[0]],
            permitted_range={feature_names[0]: [2.0, 5.0]})
        rai_insights.counterfactual.compute()

        counterfactual_data = rai_insights.counterfactual.get_data()
        self._verify_counterfactual_data(counterfactual_data)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from responsibleai import RAIInsights


class TestCausalWithDifferentDataTypes:
    def test_causal_with_object_types(self, get_adult_shap_dataset):
        data_train, data_test, treatment_features, \
            heterogeneity_features, cat_cols, \
            target_feature = get_adult_shap_dataset

        rai_i = RAIInsights(
            model=None,
            train=data_train,
            test=data_test,
            target_column=target_feature,
            task_type='classification',
            categorical_features=cat_cols,
        )

        rai_i.causal.add(
            treatment_features=treatment_features,
            heterogeneity_features=heterogeneity_features
        )

        rai_i.compute()

    def test_causal_with_categorical_types(self, get_adult_shap_dataset):
        data_train, data_test, treatment_features, \
            heterogeneity_features, cat_cols, \
            target_feature = get_adult_shap_dataset

        for c in cat_cols:
            data_train[c] = data_train[c].astype("category")
            data_test[c] = data_test[c].astype("category")

        rai_i = RAIInsights(
            model=None,
            train=data_train,
            test=data_test,
            target_column=target_feature,
            task_type='classification',
            categorical_features=cat_cols,
        )

        rai_i.causal.add(
            treatment_features=treatment_features,
            heterogeneity_features=heterogeneity_features
        )

        rai_i.compute()



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import timeit


def _measure_time(manager_compute_func):
    def compute_wrapper(*args, **kwargs):
        _separator(80)
        start_time = timeit.default_timer()
        manager_compute_func(*args, **kwargs)
        elapsed = timeit.default_timer() - start_time
        m, s = divmod(elapsed, 60)
        print('Time taken: {0} min {1} sec'.format(
              m, s))
        _separator(80)
    return compute_wrapper


def _separator(max_len):
    print('=' * max_len)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import json
import os
import shutil
import zipfile
from urllib.request import urlretrieve

import numpy as np
import pandas as pd
import pytest
# Defines common utilities for responsibleai tests
from sklearn.model_selection import train_test_split

from rai_test_utils.datasets.tabular import \
    create_iris_data as _create_iris_data
from raiutils.common.retries import retry_function


def create_iris_data():
    (
        X_train,
        X_test,
        y_train,
        y_test,
        feature_names,
        classes,
    ) = _create_iris_data()

    # create duplicate index
    test_index_list = X_test.index.tolist()
    test_index_list.append(test_index_list[0])
    X_test = X_test.reindex(test_index_list)
    y_test_list = y_test.tolist()
    y_test_list.append(y_test[0])
    y_test = np.array(y_test_list)

    return X_train, X_test, y_train, y_test, feature_names, classes


class FetchDiceAdultCensusIncomeDataset(object):
    def __init__(self):
        pass

    def fetch(self):
        """
        Loads adult income dataset from
        https://archive.ics.uci.edu/ml/datasets/Adult and prepares the
        data for data analysis based on https://rpubs.com/H_Zhu/235617

        :return adult_data: returns preprocessed adult income dataset.
        """
        # TODO: Revert to using load_adult_income_dataset once dice-ml has a
        # new release with the fix.
        # Download the adult dataset from
        # https://archive.ics.uci.edu/static/public/2/adult.zip as a zip folder
        outdirname = "adult"
        zipfilename = outdirname + ".zip"
        urlretrieve(
            "https://archive.ics.uci.edu/static/public/2/adult.zip",
            zipfilename,
        )
        with zipfile.ZipFile(zipfilename, "r") as unzip:
            unzip.extractall(outdirname)

        raw_data = np.genfromtxt(
            outdirname + "/adult.data",
            delimiter=", ",
            dtype=str,
            invalid_raise=False,
        )

        #  column names from "https://archive.ics.uci.edu/ml/datasets/Adult"
        column_names = [
            "age",
            "workclass",
            "fnlwgt",
            "education",
            "educational-num",
            "marital-status",
            "occupation",
            "relationship",
            "race",
            "gender",
            "capital-gain",
            "capital-loss",
            "hours-per-week",
            "native-country",
            "income",
        ]

        adult_data = pd.DataFrame(raw_data, columns=column_names)

        # For more details on how the below transformations are made,
        # please refer to https://rpubs.com/H_Zhu/235617
        adult_data = adult_data.astype(
            {
                "age": np.int64,
                "educational-num": np.int64,
                "hours-per-week": np.int64,
            }
        )

        adult_data = adult_data.replace(
            {
                "workclass": {
                    "Without-pay": "Other/Unknown",
                    "Never-worked": "Other/Unknown",
                }
            }
        )
        adult_data = adult_data.replace(
            {
                "workclass": {
                    "Federal-gov": "Government",
                    "State-gov": "Government",
                    "Local-gov": "Government",
                }
            }
        )
        adult_data = adult_data.replace(
            {
                "workclass": {
                    "Self-emp-not-inc": "Self-Employed",
                    "Self-emp-inc": "Self-Employed",
                }
            }
        )
        adult_data = adult_data.replace(
            {
                "workclass": {
                    "Never-worked": "Self-Employed",
                    "Without-pay": "Self-Employed",
                }
            }
        )
        adult_data = adult_data.replace({"workclass": {"?": "Other/Unknown"}})

        adult_data = adult_data.replace(
            {
                "occupation": {
                    "Adm-clerical": "White-Collar",
                    "Craft-repair": "Blue-Collar",
                    "Exec-managerial": "White-Collar",
                    "Farming-fishing": "Blue-Collar",
                    "Handlers-cleaners": "Blue-Collar",
                    "Machine-op-inspct": "Blue-Collar",
                    "Other-service": "Service",
                    "Priv-house-serv": "Service",
                    "Prof-specialty": "Professional",
                    "Protective-serv": "Service",
                    "Tech-support": "Service",
                    "Transport-moving": "Blue-Collar",
                    "Unknown": "Other/Unknown",
                    "Armed-Forces": "Other/Unknown",
                    "?": "Other/Unknown",
                }
            }
        )

        adult_data = adult_data.replace(
            {
                "marital-status": {
                    "Married-civ-spouse": "Married",
                    "Married-AF-spouse": "Married",
                    "Married-spouse-absent": "Married",
                    "Never-married": "Single",
                }
            }
        )

        adult_data = adult_data.replace(
            {
                "race": {
                    "Black": "Other",
                    "Asian-Pac-Islander": "Other",
                    "Amer-Indian-Eskimo": "Other",
                }
            }
        )

        adult_data = adult_data[
            [
                "age",
                "workclass",
                "education",
                "marital-status",
                "occupation",
                "race",
                "gender",
                "hours-per-week",
                "income",
            ]
        ]

        adult_data = adult_data.replace({"income": {"<=50K": 0, ">50K": 1}})

        adult_data = adult_data.replace(
            {
                "education": {
                    "Assoc-voc": "Assoc",
                    "Assoc-acdm": "Assoc",
                    "11th": "School",
                    "10th": "School",
                    "7th-8th": "School",
                    "9th": "School",
                    "12th": "School",
                    "5th-6th": "School",
                    "1st-4th": "School",
                    "Preschool": "School",
                }
            }
        )

        adult_data = adult_data.rename(
            columns={
                "marital-status": "marital_status",
                "hours-per-week": "hours_per_week",
            }
        )

        train, _ = train_test_split(adult_data, test_size=0.2, random_state=17)
        adult_data = train.reset_index(drop=True)

        # Remove the downloaded dataset
        if os.path.isdir(outdirname):
            entire_path = os.path.abspath(outdirname)
            shutil.rmtree(entire_path)

        return adult_data


def create_adult_income_dataset(create_small_dataset=True):
    fetcher = FetchDiceAdultCensusIncomeDataset()
    action_name = "Adult dataset download"
    err_msg = "Failed to download adult dataset"
    max_retries = 4
    retry_delay = 60
    dataset = retry_function(
        fetcher.fetch,
        action_name,
        err_msg,
        max_retries=max_retries,
        retry_delay=retry_delay,
    )
    continuous_features = ["age", "hours_per_week"]
    target_name = "income"
    target = dataset[target_name]
    classes = list(np.unique(target))
    feature_columns = dataset.drop(columns=[target_name]).columns.tolist()
    feature_range_keys = ["column_name", "range_type", "unique_values"]
    categorical_features = list(
        set(dataset.columns) - set(continuous_features) - set([target_name])
    )
    # Split data into train and test
    if create_small_dataset:
        data_train, data_test, y_train, y_test = train_test_split(
            dataset,
            target,
            train_size=500,
            test_size=50,
            random_state=7,
            stratify=target,
        )
    else:
        data_train, data_test, y_train, y_test = train_test_split(
            dataset, target, test_size=5000, random_state=7, stratify=target
        )
    return (
        data_train,
        data_test,
        y_train,
        y_test,
        categorical_features,
        continuous_features,
        target_name,
        classes,
        feature_columns,
        feature_range_keys,
    )


def assert_series_and_dict_equal(left: pd.Series, right: dict):
    left_json = left.to_json(orient="index")
    left_dict = json.loads(left_json)
    assert left_dict == pytest.approx(right)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pytest

from raiutils.exceptions import UserConfigValidationException
from responsibleai.feature_metadata import FeatureMetadata


class TestFeatureMetadata:
    def test_feature_metadata(self):
        feature_metadata = FeatureMetadata()
        assert feature_metadata.identity_feature_name is None
        assert feature_metadata.datetime_features is None
        assert feature_metadata.time_series_id_features is None
        assert feature_metadata.categorical_features is None
        assert feature_metadata.dropped_features is None

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': None,
            'datetime_features': None,
            'time_series_id_features': None,
            'categorical_features': None,
            'dropped_features': None
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_identity_feature(self):
        feature_metadata = FeatureMetadata(identity_feature_name='id')
        assert feature_metadata.identity_feature_name == 'id'
        assert feature_metadata.datetime_features is None
        assert feature_metadata.time_series_id_features is None
        assert feature_metadata.categorical_features is None
        assert feature_metadata.dropped_features is None
        with pytest.raises(
                UserConfigValidationException,
                match='The given identity feature name id is not present '
                      'in the provided features: id1, s1, s2.'):
            feature_metadata.validate_feature_metadata_with_user_features(
                feature_names=['id1', 's1', 's2'])

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': 'id',
            'datetime_features': None,
            'time_series_id_features': None,
            'categorical_features': None,
            'dropped_features': None
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_datetime_features(self):
        feature_metadata = FeatureMetadata(datetime_features=['d1'])
        assert feature_metadata.identity_feature_name is None
        assert feature_metadata.datetime_features == ['d1']
        assert feature_metadata.time_series_id_features is None
        assert feature_metadata.categorical_features is None
        assert feature_metadata.dropped_features is None

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': None,
            'datetime_features': ['d1'],
            'time_series_id_features': None,
            'categorical_features': None,
            'dropped_features': None
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_categorical_features(self):
        feature_metadata = FeatureMetadata(
            categorical_features=['c1', 'c2'])
        assert feature_metadata.identity_feature_name is None
        assert feature_metadata.datetime_features is None
        assert feature_metadata.time_series_id_features is None
        assert feature_metadata.categorical_features == ['c1', 'c2']
        assert feature_metadata.dropped_features is None

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': None,
            'datetime_features': None,
            'time_series_id_features': None,
            'categorical_features': ['c1', 'c2'],
            'dropped_features': None
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_dropped_features(self):
        feature_metadata = FeatureMetadata(dropped_features=['d1', 'd2'])
        assert feature_metadata.identity_feature_name is None
        assert feature_metadata.datetime_features is None
        assert feature_metadata.time_series_id_features is None
        assert feature_metadata.categorical_features is None
        assert feature_metadata.dropped_features == ['d1', 'd2']

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': None,
            'datetime_features': None,
            'categorical_features': None,
            'dropped_features': ['d1', 'd2'],
            'time_series_id_features': None
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_time_series_id_features(self):
        feature_metadata = FeatureMetadata(
            time_series_id_features=['g1', 'g2'])
        assert feature_metadata.identity_feature_name is None
        assert feature_metadata.datetime_features is None
        assert feature_metadata.categorical_features is None
        assert feature_metadata.dropped_features is None
        assert feature_metadata.time_series_id_features == ['g1', 'g2']
        with pytest.raises(
                UserConfigValidationException,
                match='The given time series ID feature g1 is not present '
                      'in the provided features: A, B, C, D, E, F, G.'):
            feature_metadata.validate_feature_metadata_with_user_features(
                feature_names=['A', 'B', 'C', 'D', 'E', 'F', 'G'])

        feature_metadata_dict = feature_metadata.to_dict()
        expected_feature_metadata_dict = {
            'identity_feature_name': None,
            'datetime_features': None,
            'categorical_features': None,
            'dropped_features': None,
            'time_series_id_features': ['g1', 'g2']
        }
        assert feature_metadata_dict == expected_feature_metadata_dict

    def test_feature_metadata_with_other_feature_metadata(self):
        feature_metadata_1 = FeatureMetadata(
            identity_feature_name='id',
            datetime_features=['d1'],
            categorical_features=['c1', 'c2'],
            dropped_features=['d3', 'd4'])

        feature_metadata_2 = FeatureMetadata(
            identity_feature_name='id',
            datetime_features=['d1'],
            categorical_features=['c1', 'c2'],
            dropped_features=['d3', 'd4'])

        feature_metadata_3 = FeatureMetadata(
            identity_feature_name='id',
            datetime_features=['d1'],
            categorical_features=['c1', 'c2'],
            dropped_features=['d3', 'd5'])

        assert feature_metadata_1 == feature_metadata_2
        assert feature_metadata_1 != feature_metadata_3



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.
"""Result of causal analysis."""

import json
import re
import uuid
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

import pandas as pd
import semver

from responsibleai._interfaces import CausalConfig as CausalConfigInterface
from responsibleai._interfaces import (CausalData, CausalPolicy,
                                       CausalPolicyGains,
                                       CausalPolicyTreeInternal,
                                       CausalPolicyTreeLeaf, ComparisonTypes)
from responsibleai._internal.constants import (FileFormats,
                                               SerializationAttributes)
from responsibleai._tools.causal.causal_config import CausalConfig
from responsibleai._tools.causal.causal_constants import ResultAttributes
from responsibleai._tools.shared.base_result import BaseResult
from responsibleai._tools.shared.versions import CausalVersions


class CausalResult(BaseResult['CausalResult']):
    """Result of causal analysis."""

    _ATTRIBUTES = [
        'config',
        'causal_analysis',
        'global_effects',
        'local_effects',
        'policies',
    ]

    def __init__(
        self,
        config: Optional[CausalConfig] = None,
        causal_analysis: Optional[Any] = None,
        global_effects: Optional[pd.DataFrame] = None,
        local_effects: Optional[pd.DataFrame] = None,
        policies: Optional[Any] = None,
    ):
        super().__init__(str(uuid.uuid4()), CausalVersions.get_current())

        self.config = config

        self.causal_analysis = causal_analysis
        self.global_effects = global_effects
        self.local_effects = local_effects
        self.policies = policies

    @property
    def is_computed(self):
        """Check if the causal analysis was performed.

        :return: True if causal analysis was performed and False
            otherwise.
        :rtype: boolean
        """
        return self.causal_analysis is not None or \
            self.global_effects is not None or \
            self.local_effects is not None or \
            self.policies is not None

    def _create_causal_data_object(self):
        """Create a causal data object.

        :return: An object of type CausalData.
        :rtype: CausalData
        """
        causal_data = CausalData()

        causal_data.id = self.id
        causal_data.version = self.version
        causal_data.config = self._get_config_object(self.config)

        return causal_data

    def _get_dashboard_object(self):
        causal_data = self._create_causal_data_object()

        causal_data.global_effects = self.global_effects\
            .reset_index().to_dict(orient='records')
        local_dicts = self.local_effects.groupby('sample').apply(
            lambda x: x.reset_index().to_dict(
                orient='records')).values
        causal_data.local_effects = [list(v) for v in local_dicts]
        causal_data.policies = [self._get_policy_object(p) for p in
                                self.policies]

        return causal_data

    def _whatif(self, X, X_feature_new, feature_name, y, alpha=0.1):
        return self.causal_analysis.whatif(
            X, X_feature_new, feature_name, y, alpha=alpha)

    def _global_cohort_effects(self, X_test):
        """Get global causal effects for cohort data.

        :param X_test: The data for which the causal policy
                       needs to be generated.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects.
        :rtype: CausalData
        """
        causal_data = self._create_causal_data_object()
        causal_data.global_effects = \
            self.causal_analysis.cohort_causal_effect(
                X_test, alpha=self.config.alpha,
                keep_all_levels=True).reset_index().to_dict(
                    orient="records")
        return causal_data

    def _local_instance_effects(self, X_test):
        """Get local causal effects for a given data point.

        :param X_test: The data for which the local causal effects
                       needs to be generated for a given point.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects for a given point.
        :rtype: CausalData
        """
        causal_data = self._create_causal_data_object()

        local_effects = self.causal_analysis.local_causal_effect(
            X_test, alpha=self.config.alpha,
            keep_all_levels=True)

        local_dicts = local_effects.groupby('sample').apply(
            lambda x: x.reset_index().to_dict(
                orient='records')).values
        causal_data.local_effects = [list(v) for v in local_dicts]

        return causal_data

    def _global_cohort_policy(self, X_test):
        """Get global causal policy for cohort data.

        :param X_test: The data for which the causal policy
                       needs to be generated.
        :type X_test: Any
        :return: An object of type CausalData with
                 causal effects.
        :rtype: CausalData
        """
        causal_data = self._create_causal_data_object()

        causal_config = self.config
        if isinstance(causal_config.treatment_cost, int) and \
                causal_config.treatment_cost == 0:
            revised_treatment_cost = [0] * len(
                causal_config.treatment_features)
        else:
            revised_treatment_cost = causal_config.treatment_cost

        policies = []
        for i in range(len(causal_config.treatment_features)):
            policy = self._create_policy(
                X_test,
                causal_config.treatment_features[i],
                revised_treatment_cost[i],
                causal_config.alpha, causal_config.max_tree_depth,
                causal_config.min_tree_leaf_samples)
            policies.append(policy)
        causal_data.policies = [self._get_policy_object(p) for p in
                                policies]
        return causal_data

    def _create_policy(
        self,
        X_test,
        treatment_feature,
        treatment_cost,
        alpha,
        max_tree_depth,
        min_tree_leaf_samples,
    ):
        local_policies = self.causal_analysis.individualized_policy(
            X_test, treatment_feature,
            treatment_costs=treatment_cost,
            alpha=alpha)

        tree = self.causal_analysis._policy_tree_output(
            X_test, treatment_feature,
            treatment_costs=treatment_cost,
            max_depth=max_tree_depth,
            min_samples_leaf=min_tree_leaf_samples,
            alpha=alpha)

        return {
            ResultAttributes.TREATMENT_FEATURE: treatment_feature,
            ResultAttributes.CONTROL_TREATMENT: tree.control_name,
            ResultAttributes.LOCAL_POLICIES: local_policies,
            ResultAttributes.POLICY_GAINS: {
                ResultAttributes.RECOMMENDED_POLICY_GAINS:
                    tree.policy_value,
                ResultAttributes.TREATMENT_GAINS: tree.always_treat,
            },
            ResultAttributes.POLICY_TREE: tree.tree_dictionary
        }

    def _get_config_object(self, config):
        config_object = CausalConfigInterface()
        config_object.treatment_features = config.treatment_features
        return config_object

    def _get_policy_object(self, policy):
        policy_object = CausalPolicy()
        policy_object.treatment_feature = policy[
            ResultAttributes.TREATMENT_FEATURE]
        policy_object.control_treatment = policy[
            ResultAttributes.CONTROL_TREATMENT]
        policy_object.local_policies = self._get_local_policies_object(
            policy[ResultAttributes.LOCAL_POLICIES])
        policy_object.policy_gains = self._get_policy_gains_object(
            policy[ResultAttributes.POLICY_GAINS])
        policy_object.policy_tree = self._get_policy_tree_object(
            policy[ResultAttributes.POLICY_TREE])
        return policy_object

    def _get_local_policies_object(self, local_policies):
        if local_policies is None:
            return None

        return local_policies.reset_index().\
            to_dict(orient='records')

    def _get_policy_gains_object(self, policy_gains):
        policy_gains_object = CausalPolicyGains()
        policy_gains_object.recommended_policy_gains = \
            policy_gains[ResultAttributes.RECOMMENDED_POLICY_GAINS]
        policy_gains_object.treatment_gains = \
            policy_gains[ResultAttributes.TREATMENT_GAINS]
        return policy_gains_object

    def _get_policy_tree_object(self, policy_tree):
        if policy_tree[ResultAttributes.LEAF]:
            policy_tree_object = CausalPolicyTreeLeaf()
            policy_tree_object.leaf = policy_tree[ResultAttributes.LEAF]
            policy_tree_object.n_samples = policy_tree[
                ResultAttributes.N_SAMPLES]
            policy_tree_object.treatment = policy_tree[
                ResultAttributes.TREATMENT]
        else:
            policy_tree_object = CausalPolicyTreeInternal()
            policy_tree_object.leaf = policy_tree[ResultAttributes.LEAF]
            feature, comparison, value = self._parse_comparison(
                policy_tree, self.config.categorical_features)
            policy_tree_object.feature = feature
            policy_tree_object.right_comparison = comparison
            policy_tree_object.comparison_value = value
            policy_tree_object.left = self._get_policy_tree_object(
                policy_tree[ResultAttributes.LEFT])
            policy_tree_object.right = self._get_policy_tree_object(
                policy_tree[ResultAttributes.RIGHT])
        return policy_tree_object

    @classmethod
    def _parse_comparison(
        cls,
        policy_tree: Any,
        categoricals: List[str],
    ) -> Tuple[str, str, Union[float, int, str]]:
        """Attempt to parse a categorical comparison from a policy tree node.

        The default assumption is that the feature is continuous and will
        have a real-valued threshold.

        This function checks known categorical features and attempts to parse
        a (feature, category) pair that would come from a one-hot encoding.

        Example (continuous):
            Original feature: "Size"
            Original comparison value: 0.891
            Parsed feature: "Size"
            Parsed comparison: "gt"
            Parsed comparison value: 0.891

        Example (categorical):
            Original feature: "Fruit_apple"
            Original comparison value: 0.5
            Parsed feature: "Fruit"
            Parsed comparison: "eq"
            Parsed comparison value: "apple"

        """
        tree_feature = policy_tree[ResultAttributes.FEATURE]
        threshold = policy_tree[ResultAttributes.THRESHOLD]

        # Default assumption is a continuous feature
        feature = tree_feature
        comparison = ComparisonTypes.GT
        value = threshold

        # Check for a categorical feature
        for cat_feature in sorted(categoricals)[::-1]:
            pattern = f'({cat_feature})_(.+)'
            match = re.match(pattern, tree_feature)
            if match is not None:
                feature = match.group(1)
                value = match.group(2)
                comparison = ComparisonTypes.EQ
                break

        return feature, comparison, value

    @classmethod
    def _get_schema(cls, version: str):
        cls._validate_version(version)

        schema_directory = Path(__file__).parent / \
            SerializationAttributes.DASHBOARD_SCHEMAS
        schema_filename = f'schema_{version}{FileFormats.JSON}'
        schema_filepath = schema_directory / schema_filename
        with open(schema_filepath, 'r') as f:
            return json.load(f)

    @classmethod
    def _validate_version(cls, version):
        semver_version = semver.VersionInfo.parse(version)
        if semver_version.compare(CausalVersions.get_current()) > 0:
            raise ValueError("The installed version of responsibleai "
                             "is not compatible with causal result version "
                             f"{version}. Please upgrade in order to load "
                             "this result.")

        if version not in CausalVersions.get_all():
            raise ValueError(f"Invalid version for causal result: {version}")

    def _get_attributes(self):
        return self._ATTRIBUTES



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from abc import ABC, abstractmethod
from typing import List

import pandas as pd


class BalanceMeasures(ABC):
    """
    An abstract class that computes measures for a given dataset
    and given columns of interest.
    """

    def __init__(self, cols_of_interest: List[str]) -> None:
        """
        Construct a BalanceMeasures class for calculating
        measures on the given columns of interest.

        :param cols_of_interest: The list of columns to compute measures on.
        :type cols_of_interest: List[str]
        """
        self.cols_of_interest = cols_of_interest

    @abstractmethod
    def measures(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Returns measures for the given dataset and columns of interest.

        :param dataset: The dataset to compute measures on.
        :type dataset: pd.DataFrame
        :return: A pandas DataFrame containing the measures.
        :rtype: pd.DataFrame
        """
        ...



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from typing import Callable, Dict, List

import numpy as np
import pandas as pd

from responsibleai.databalanceanalysis.balance_measures import BalanceMeasures
from responsibleai.databalanceanalysis.constants import Measures


def _get_generalized_entropy_index(
    benefits: np.array,
    alpha: float,
    use_abs_val: bool,
    error_tolerance: float = 1e-12,
) -> float:
    """
    Returns the general entropy index, a measure of redundancy in the data.
    https://en.wikipedia.org/wiki/Generalized_entropy_index

    :param benefits: The probabilities of each feature value combination
        occurring in the data.
    :type benefits: np.array
    :param alpha: A parameter which regulates the weight given to distances
        between benefits at different parts of the distribution.
    :type alpha: float
    :param use_abs_val: Whether to use the absolute value of the benefits
        or not.
    :type use_abs_val: bool
    :param error_tolerance: The error tolerance to use when calculating
        the generalized entropy index.
    :type error_tolerance: float
    :return: The generalized entropy index.
    :rtype: float
    """
    if use_abs_val:
        benefits = np.absolute(benefits)

    benefits_mean = np.mean(benefits)
    norm_benefits = benefits / benefits_mean
    count = norm_benefits.size

    if abs(alpha - 1.0) < error_tolerance:
        gei = np.sum(norm_benefits * np.log(norm_benefits)) / count
    elif abs(alpha) < error_tolerance:
        gei = np.sum(-np.log(norm_benefits)) / count
    else:
        gei = np.sum(np.power(norm_benefits, alpha) - 1.0) / (
            count * alpha * (alpha - 1.0)
        )

    return gei


def get_atkinson_index(
    benefits: np.array,
    epsilon: float = 1.0,
    error_tolerance: float = 1e-12,
) -> float:
    """
    Returns the Atkinson Index, a measure of income inequality that
    indicates the percentage of the data that needs to be foregone
    to have equal shares of income.
    https://en.wikipedia.org/wiki/Atkinson_index

    :param benefits: The probabilities of each feature value combination
        occurring in the data.
    :type benefits: np.array
    :param epsilon: A parameter that regulates the weight given
        to the distances between benefits.
    :type epsilon: float
    :param error_tolerance: The error tolerance to use when calculating
        the Atkinson Index.
    :type error_tolerance: float
    :return: The Atkinson Index.
    :rtype: float
    """
    count = benefits.size
    benefits_mean = np.mean(benefits)
    norm_benefits = benefits / benefits_mean
    alpha = 1 - epsilon

    if abs(alpha) < error_tolerance:
        ati = 1.0 - np.power(np.prod(norm_benefits), 1.0 / count)
    else:
        power_mean = np.sum(np.power(norm_benefits, alpha)) / count
        ati = 1.0 - np.power(power_mean, 1.0 / alpha)

    return ati


def get_theil_t_index(benefits: np.array) -> float:
    """
    Returns Theil T Index, a measure of income inequality that is more
    sensitive to differences at the top of the distribution.
    https://en.wikipedia.org/wiki/Theil_index

    :param benefits: The probabilities of each feature value combination
        occurring in the data.
    :type benefits: np.array
    :return: The Theil T Index.
    :rtype: float
    """
    return _get_generalized_entropy_index(
        benefits=benefits, alpha=1.0, use_abs_val=True
    )


def get_theil_l_index(benefits: np.array) -> float:
    """
    Returns Theil L Index, a measure of income inequality that is
    more sensitive to differences at the lower end of the distribution.
    https://en.wikipedia.org/wiki/Theil_index

    :param benefits: The probabilities of each feature value combination
        occurring in the data.
    :type benefits: np.array
    :return: The Theil L Index.
    :rtype: float
    """
    return _get_generalized_entropy_index(
        benefits=benefits, alpha=0.0, use_abs_val=True
    )


class AggregateBalanceMeasures(BalanceMeasures):
    """
    This class computes a set of aggregated balance measures that represents
    how balanced the given dataset is along the given columns of interest.
    """

    AGGREGATE_METRICS: Dict[Measures, Callable[[np.array], float]] = {
        Measures.THEIL_L_INDEX: get_theil_l_index,
        Measures.THEIL_T_INDEX: get_theil_t_index,
        Measures.ATKINSON_INDEX: get_atkinson_index,
    }

    def __init__(self, cols_of_interest: List[str]) -> None:
        """
        Construct a AggregateBalanceMeasures class for calculating
        aggregate balance measures on a dataset.

        :param cols_of_interest: The list of columns to compute
            aggregate balance measures on.
        :type cols_of_interest: List[str]
        """
        super().__init__(cols_of_interest=cols_of_interest)

    def measures(self, dataset: pd.DataFrame) -> pd.DataFrame:
        """
        Returns aggregate balance measures for the given dataset and
        columns of interest.

        The following measures are computed:
        * Atkinson Index - https://en.wikipedia.org/wiki/Atkinson_index
        * Theil Index (L and T) - https://en.wikipedia.org/wiki/Theil_index

        :param dataset: The dataset to compute aggregate measures on.
        :type dataset: pd.DataFrame
        :return: A pandas DataFrame that contains the names and values of
            aggregate balance measures for the given dataset and
            columns of interest.
        :rtype: pd.DataFrame
        """
        aggregate_measures_dict = {}
        benefits = (
            dataset.groupby(self.cols_of_interest).size() / dataset.shape[0]
        )

        for measure, func in self.AGGREGATE_METRICS.items():
            aggregate_measures_dict[measure.value] = [func(benefits)]

        aggregate_measures = pd.DataFrame.from_dict(aggregate_measures_dict)
        return aggregate_measures



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import pandas as pd

from responsibleai.databalanceanalysis import DistributionBalanceMeasures

from ..common_utils import assert_series_and_dict_equal
from .conftest import SYNTHETIC_DATA_ETHNICITY, SYNTHETIC_DATA_GENDER


class TestDistributionBalanceMeasures:
    def test_feature_1_synthetic_data(
        self, synthetic_data, expected_distribution_measures_gender
    ):
        dist_measures = (
            DistributionBalanceMeasures(
                cols_of_interest=[SYNTHETIC_DATA_GENDER]
            )
            .measures(dataset=synthetic_data)
            .iloc[0]
        )
        assert_series_and_dict_equal(
            dist_measures, expected_distribution_measures_gender
        )

    def test_feature_2_synthetic_data(
        self, synthetic_data, expected_distribution_measures_ethnicity
    ):
        dist_measures = (
            DistributionBalanceMeasures(
                cols_of_interest=[SYNTHETIC_DATA_ETHNICITY]
            )
            .measures(dataset=synthetic_data)
            .iloc[0]
        )
        assert_series_and_dict_equal(
            dist_measures, expected_distribution_measures_ethnicity
        )

    def test_adult_data(
        self, adult_data, adult_data_distribution_balance_measures
    ):
        train_df, test_df, _, _, _ = adult_data
        full_df = pd.concat([train_df, test_df])
        actual = DistributionBalanceMeasures(
            cols_of_interest=["race", "gender"]
        ).measures(dataset=full_df)
        pd.testing.assert_frame_equal(
            actual.reset_index(drop=True),
            adult_data_distribution_balance_measures.reset_index(drop=True),
        )
